use ufbx;

fn create_vertex_buffer<T>(vertices: &[T]) {
    println!(".. {} vertices", vertices.len());
}

fn create_index_buffer(indices: &[u32]) {
    println!(".. {} indices", indices.len());
}

// GPU vertex representation.
// In practice you would need to use more compact custom vector types as
// by default `ufbx::Real` is 64 bits wide.
#[repr(C)]
#[derive(Clone, Copy)]
struct Vertex {
    position: ufbx::Vec3,
    normal: ufbx::Vec3,
    uv: ufbx::Vec2,
}

fn convert_mesh_part(mesh: &ufbx::Mesh, part: &ufbx::MeshPart) {
    let mut vertices: Vec<Vertex> = Vec::new();
    let mut tri_indices = vec![0u32; mesh.max_face_triangles * 3];

    // Iterate over each face using the specific material.
    for &face_index in &part.face_indices {
        let face = mesh.faces[face_index as usize];

        // Triangulate the face into `tri_indices[]`.
        let num_tris = mesh.triangulate_face(&mut tri_indices, face);
        let num_tri_corners = (num_tris * 3) as usize;

        // Iterate over each triangle corner contiguously.
        for &index in &tri_indices[..num_tri_corners] {
            vertices.push(Vertex {
                position: mesh.vertex_position[index as usize],
                normal: mesh.vertex_normal[index as usize],
                uv: mesh.vertex_uv[index as usize],
            });
        }
    }

    // Should have written all the vertices.
    assert!(vertices.len() == part.num_triangles * 3);

    // Generate the index buffer.
    let mut streams = [
        ufbx::VertexStream::new(&mut vertices),
    ];
    let mut indices = vec![0u32; part.num_triangles * 3];

    // This call will deduplicate vertices, modifying the arrays passed in `streams[]`,
    // indices are written in `indices[]` and the number of unique vertices is returned.
    let num_vertices = ufbx::generate_indices(
            &mut streams, &mut indices, ufbx::AllocatorOpts::default())
        .expect("expected to generate indices");

    // Trim to only unique vertices.
    vertices.truncate(num_vertices);

    create_vertex_buffer(&vertices);
    create_index_buffer(&indices);
}

fn main() {
    let scene = ufbx::load_file("material_parts.fbx", Default::default()).expect("failed to load scene");

    for mesh in &scene.meshes {
        println!("mesh '{}'", mesh.element.name);

        for part in &mesh.material_parts {
            let material = mesh.materials.get(part.index as usize);
            let material_name = material
                .map(|m| m.element.name.as_ref())
                .unwrap_or("");

            println!(". [{}] material: {}", part.index, material_name);
            convert_mesh_part(mesh, part);
        }
    }
}
